import ctypes

from glob import glob
from typing import cast, List, Optional, Self, Tuple

from edge import eventlog
from edge.types.image import ImRgba


class VsServiceArgs:
    def __init__(self, store: int, chan: str = '0', log_level: str = 'info'):
        self.store = store
        self.chan = chan
        self.log_level = log_level


_IMAGE_POOL_MAX = 1


class VsImageInfo:
    def __init__(self, ts: int):
        self._ts = ts

    @property
    def ts(self) -> int:
        """
        :returns:
            value in usec
        """
        return self._ts


class VsError(RuntimeError):
    """
    Generic Error.
    """
    pass


class VsImageBufferError(VsError):
    def __init__(self, min_w: int, min_h: int):
        self.min_w = min_w
        self.min_h = min_h


class VsSessionAbsentError(VsError):
    """
    Session is absent in the pool.
    """
    pass


class VsSessionExpiredError(VsError):
    """
    Session is expired.
    Session should be closed and opened again.
    """
    pass


class IVsService:
    def __init__(self):
        pass

    def export_image_rgba(self, im_rgba: ImRgba, im_infos: List[VsImageInfo]) -> Optional[VsImageInfo]:
        """
        :raise VsImageBufferError:
        :raise VsError:
        """
        raise NotImplementedError

    def search_asc(self, min_ts: int, max_ts: int, ts_buffer: List[int]) -> None:
        raise NotImplementedError

    def export_image_rgba2(self, im_rgba: ImRgba, im_infos: List[VsImageInfo]) -> Optional[VsImageInfo]:
        """
        :raise VsError:
        """
        while True:
            try:
                answer = self.export_image_rgba(im_rgba, im_infos)
                return answer
            except VsImageBufferError as x:
                min_w, min_h = x.min_w, x.min_h

            im_rgba.reset(min_w, min_h, im_rgba.dh, im_rgba.ts)


def init_service(args: VsServiceArgs) -> IVsService:
    return _VsService(args)


def fini_service(service: IVsService):
    self = cast(_VsService, service)
    self.destroy()


class _Api:
    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2
    RC_CONN = -3

    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    IMAGE_GRAY = 1
    IMAGE_RGBA = 2
    IMAGE_JPEG = 3

    VIDEO_H264_AVI = 1
    VIDEO_H264_MKV = 2
    VIDEO_H264_MP4 = 3

    _lib = ctypes.CDLL(glob('/opt/edge/lib/libblobstore-bsv*.so')[0])

    # int
    # BsvOpen(uint16_t aStore, uint8_t aLogLevel, void **aSession);

    open = _lib.BsvOpen
    open.argtypes = (ctypes.c_uint16, ctypes.c_uint8, ctypes.POINTER(ctypes.c_voidp))
    open.restype = ctypes.c_int

    # int
    # BsvLoadImage(void *aSession, const char *aChannel, uint64_t anImTs, uint8_t anImFormat, uint8_t *anImData,
    #     uint32_t *anImSize,uint16_t *anImWidth, uint16_t *anImHeight);

    load = _lib.BsvLoadImage
    load.argtypes = (
        ctypes.c_voidp, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8),
        ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint16), ctypes.POINTER(ctypes.c_uint16))

    load.restype = ctypes.c_int

    # int
    # BsvSearchTsAsc (void *aSession, const char *aChannel, uint64_t aMinTs, uint64_t aMaxTs,
    #   uint64_t *aTsList, uint16_t *aTsCount);

    search_asc = _lib.BsvSearchTsAsc
    search_asc.argtypes = (
        ctypes.c_voidp, ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(ctypes.c_uint64), ctypes.POINTER(ctypes.c_uint16)
    )

    search_asc.restype = ctypes.c_int

    # void
    # BsvClose(void *aSession);
    close = _lib.BsvClose
    close.argtype = ctypes.c_voidp

    def __init__(self):
        pass


class _VsSession:
    def __init__(self, handle, chan: str):
        """
        :type handle: ctypes.c_void_p | None
        :type chan: str
        """
        self._chan = chan
        self._handle = handle
        self.closed = handle is None

    @staticmethod
    def open(store: int, chan: str, log_level: int = _Api.LOG_I) -> Self:
        """
        :raise VsSessionExpiredError:
        """
        handle = ctypes.c_void_p(0)
        rc = _Api.open(store, log_level, handle)

        if rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvOpen (rc:%s)' % rc)
            raise VsSessionExpiredError

        return _VsSession(handle=handle, chan=chan)

    def export_image_rgba(self, im_ts: int, im_buffer: str) -> Tuple[int, int]:
        """
        :param im_ts:
            value in msec
        :param im_buffer

        :raise VsSessionExpiredError:
        :raise VsError:
        """
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        im_data = ctypes.cast(im_buffer, ctypes.POINTER(ctypes.c_uint8))
        im_size = ctypes.c_uint32(ctypes.sizeof(im_buffer))
        im_w, im_h = ctypes.c_uint16(), ctypes.c_uint16()

        rc = _Api.load(
            self._handle, self._chan, ctypes.c_uint64(im_ts), ctypes.c_uint8(_Api.IMAGE_RGBA),
            im_data, im_size, im_w, im_h)

        if im_size.value > ctypes.sizeof(im_buffer):
            eventlog.log_e('fail: buffer too small')
            return im_w.value, im_h.value

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvLoadImage (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvLoadImage (rc:%s)' % rc)
            raise VsError

        return im_w.value, im_h.value

    def search_asc(self, min_ts: int, max_ts: int, ts_buffer: List[int]) -> None:
        if self.closed:
            eventlog.log_e('fail: session closed')
            raise VsSessionExpiredError

        # ts_buffer = [1, 2, 3, 4, 5]
        # buffer_data = ctypes.cast(ts_buffer, ctypes.POINTER(ctypes.c_uint64))

        buffer_data = (ctypes.c_uint64 * len(ts_buffer))(*ts_buffer)
        buffer_size = ctypes.c_uint16(len(ts_buffer))

        rc = _Api.search_asc(
            self._handle, self._chan, ctypes.c_uint64(min_ts), ctypes.c_uint64(max_ts), buffer_data, buffer_size)

        if rc == _Api.RC_CONN:
            self.close()
            eventlog.log_e('fail: BsvSearchTsAsc (kBS_CONN)')
            raise VsSessionExpiredError

        elif rc != _Api.RC_DONE:
            eventlog.log_e('fail: BsvSearchTsAsc (rc:%s)' % rc)
            raise VsError

        print(buffer_data)
        eventlog.log_i("size {} {}".format(buffer_size.value, buffer_data))

    def close(self) -> None:
        if not self.closed:
            _Api.close(self._handle)
            self.closed = True


class _VsSessionPool:
    _LOG_NAME_TO_CODE_MAP = {
        'none': _Api.LOG_N,
        'error': _Api.LOG_E,
        'warn': _Api.LOG_W,
        'info': _Api.LOG_I,
        'debug': _Api.LOG_D,
        'trace': _Api.LOG_T
    }

    def __init__(self, bs_store: int, bs_chan: str, bs_log: str, max_count: int):
        self.max_count = max_count
        log_code = _VsSessionPool._LOG_NAME_TO_CODE_MAP.get(bs_log, 'info')
        self._conf = bs_store, bs_chan, log_code
        self._sessions = [_VsSession(handle=None, chan=bs_chan) for _ in range(max_count)]

    def __del__(self):
        """
        :rtype: None
        """
        list([x.close() for x in self._sessions])

    def get_session(self) -> _VsSession:
        """
        :raise VsSessionAbsentError:
        :raise VsError:
        """
        if not self._sessions:
            eventlog.log_e('fail: session absent')
            raise VsSessionAbsentError

        session = self._sessions.pop()
        try:
            attempt = 0
            while attempt < 3:
                attempt += 1
                try:
                    if session.closed:
                        session = _VsSession.open(*self._conf)

                    eventlog.log_d('done: session:%s' % id(session))
                    return session

                except VsSessionExpiredError:
                    session.close()

                # todo: consider to sleep a little bit
                eventlog.log_e('fail: open session (attempt:%s/3)' % attempt)

            raise VsError

        except VsError:
            self._sessions.append(session)
            raise

        except Exception:
            self._sessions.append(session)
            raise

    def put_session(self, session: _VsSession) -> None:
        assert session not in self._sessions
        self._sessions.append(session)


class _VsService(IVsService):
    def __init__(self, args: VsServiceArgs):
        IVsService.__init__(self)

        self._image_pool = _VsSessionPool(args.store, args.chan, args.log_level, _IMAGE_POOL_MAX)

    def destroy(self) -> None:
        del self._image_pool

    def export_image_rgba(self, im_rgba: ImRgba, im_infos: List[VsImageInfo]) -> Optional[VsImageInfo]:
        """
        :raise VsImageBufferError:
        :raise VsError:
        """
        session = self._image_pool.get_session()
        try:
            for im_info in im_infos:
                try:
                    image_w, image_h = session.export_image_rgba(im_info.ts // 1000, im_rgba.buffer)
                except VsSessionExpiredError:
                    raise

                except VsError:
                    eventlog.log_i('fail: export image-ts:%s (trying the next one)' % im_info.ts)
                    continue

                if image_w > im_rgba.w or image_h > im_rgba.h:
                    raise VsImageBufferError(image_w, image_h)

                elif image_w > 0 and image_h > 0:
                    im_rgba.reset(image_w, image_h, im_rgba.dh, im_info.ts)
                    return im_info

                else:
                    eventlog.log_e('fail: invalid image-wh:%sx%s' % (image_w, image_h))
                    raise VsError
        finally:
            self._image_pool.put_session(session)

    def search_asc(self, min_ts: int, max_ts: int, ts_buffer: List[int]):
        session = self._image_pool.get_session()
        try:
            try:
                session.search_asc(min_ts, max_ts, ts_buffer)
            except VsSessionExpiredError:
                raise

            except VsError:
                eventlog.log_i('fail:')
        finally:
            self._image_pool.put_session(session)
